# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import sciline
import scipp as sc
from scipp.testing import assert_allclose

import ess.polarization as pol
from ess.polarization import (
    CorrectionWorkflow,
    He3CellWorkflow,
    PolarizationAnalysisWorkflow,
    SupermirrorWorkflow,
)
from ess.polarization.correction import (
    FlipperEfficiency,
    compute_polarizing_element_correction,
)
from ess.polarization.types import (
    Analyzer,
    Down,
    HalfPolarizedCorrectedData,
    PolarizationCorrectedData,
    Polarizer,
    ReducedSampleDataBySpinChannel,
    TotalPolarizationCorrectedData,
    TransmissionFunction,
    Up,
)


class SimpleTransmissionFunction:
    def __call__(
        self, time: sc.Variable, wavelength: sc.Variable, plus_minus: str
    ) -> sc.Variable:
        if plus_minus == 'plus':
            return 10 * time * (2 + wavelength)
        else:
            return 10 * time * (2 - wavelength)

    def apply(self, da: sc.DataArray, plus_minus: str) -> float:
        time = da.coords['time']
        wavelength = da.coords['wavelength']
        return self(time, wavelength, plus_minus)


def test_compute_polarizing_element_correction() -> None:
    time = sc.linspace('event', 1, 10, 10, unit='')
    wavelength = sc.linspace('event', 0.1, 1, 10, unit='')
    events = sc.DataArray(
        sc.arange('event', 10),
        coords={'time': time, 'wavelength': wavelength},
    )
    transmission = SimpleTransmissionFunction()

    result = compute_polarizing_element_correction(
        channel=events, transmission=transmission
    )
    diag = result.diag
    off_diag = result.off_diag
    assert diag.sizes == {'event': 10}
    assert off_diag.sizes == {'event': 10}
    transmission_plus = transmission(time, wavelength, 'plus')
    transmission_minus = transmission(time, wavelength, 'minus')
    denom = transmission_plus**2 - transmission_minus**2
    assert_allclose(diag, transmission_plus / denom)
    assert_allclose(off_diag, -transmission_minus / denom)


class FakeTransmissionFunction:
    def __init__(self, coeffs: np.ndarray) -> None:
        self.coeffs = coeffs

    def apply(self, _: sc.DataArray, plus_minus: str) -> float:
        if plus_minus == 'plus':
            return sc.scalar(self.coeffs[0][0])
        else:
            return sc.scalar(self.coeffs[0][1])


def test_correction_workflow_computes_and_applies_matrix_inverse() -> None:
    ground_truth = np.array([7.0, 11.0, 13.0, 17.0])
    analyzer = np.array([[1.3, 0.9], [0.9, 1.3]])
    polarizer = np.array([[1.1, 0.7], [0.7, 1.1]])
    identity = np.array([[1.0, 0.0], [0.0, 1.0]])
    intensity = (
        np.kron(identity, analyzer) @ np.kron(polarizer, identity) @ ground_truth
    )

    workflow = CorrectionWorkflow()
    workflow[TransmissionFunction[Analyzer]] = FakeTransmissionFunction(analyzer)
    workflow[TransmissionFunction[Polarizer]] = FakeTransmissionFunction(polarizer)
    workflow[ReducedSampleDataBySpinChannel[Up, Up]] = intensity[0]
    workflow[ReducedSampleDataBySpinChannel[Up, Down]] = intensity[1]
    workflow[ReducedSampleDataBySpinChannel[Down, Up]] = intensity[2]
    workflow[ReducedSampleDataBySpinChannel[Down, Down]] = intensity[3]

    result = np.zeros(4)
    for pola in [Up, Down]:
        for ana in [Up, Down]:
            contrib = workflow.compute(PolarizationCorrectedData[pola, ana])
            contrib = sc.concat(
                [contrib.upup, contrib.updown, contrib.downup, contrib.downdown],
                'dummy',
            )
            result += contrib.values
    np.testing.assert_allclose(result, ground_truth)


@pytest.mark.parametrize('f', [0.1, 0.5, 0.9, 0.99, 1.0])
def test_workflow_with_analyzer_flipper_computes_and_applies_matrix_inverse(
    f: float,
) -> None:
    ground_truth = np.array([7.0, 11.0, 13.0, 17.0])
    analyzer = np.array([[1.3, 0.9], [0.9, 1.3]])
    polarizer = np.array([[1.1, 0.7], [0.7, 1.1]])
    identity = np.array([[1.0, 0.0], [0.0, 1.0]])

    flipper = np.array([[1.0, 0.0], [1 - f, f]])
    intensity = (
        np.kron(identity, analyzer @ flipper)
        @ np.kron(polarizer, identity)
        @ ground_truth
    )

    workflow = CorrectionWorkflow()
    workflow[TransmissionFunction[Analyzer]] = FakeTransmissionFunction(analyzer)
    workflow[TransmissionFunction[Polarizer]] = FakeTransmissionFunction(polarizer)
    workflow[ReducedSampleDataBySpinChannel[Up, Up]] = intensity[0]
    workflow[ReducedSampleDataBySpinChannel[Up, Down]] = intensity[1]
    workflow[ReducedSampleDataBySpinChannel[Down, Up]] = intensity[2]
    workflow[ReducedSampleDataBySpinChannel[Down, Down]] = intensity[3]
    workflow[FlipperEfficiency[Analyzer]] = FlipperEfficiency(f)

    result = np.zeros(4)
    for pola in [Up, Down]:
        for ana in [Up, Down]:
            contrib = workflow.compute(PolarizationCorrectedData[pola, ana])
            contrib = sc.concat(
                [contrib.upup, contrib.updown, contrib.downup, contrib.downdown],
                'dummy',
            )
            result += contrib.values
    np.testing.assert_allclose(result, ground_truth)


@pytest.mark.parametrize('f', [0.1, 0.5, 0.9, 0.99, 1.0])
def test_workflow_with_polarizer_flipper_computes_and_applies_matrix_inverse(
    f: float,
) -> None:
    ground_truth = np.array([7.0, 11.0, 13.0, 17.0])
    analyzer = np.array([[1.3, 0.9], [0.9, 1.3]])
    polarizer = np.array([[1.1, 0.7], [0.7, 1.1]])
    identity = np.array([[1.0, 0.0], [0.0, 1.0]])

    flipper = np.array([[1.0, 0.0], [1 - f, f]])
    intensity = (
        np.kron(identity, analyzer)
        @ np.kron(flipper @ polarizer, identity)
        @ ground_truth
    )

    workflow = CorrectionWorkflow()
    workflow[TransmissionFunction[Analyzer]] = FakeTransmissionFunction(analyzer)
    workflow[TransmissionFunction[Polarizer]] = FakeTransmissionFunction(polarizer)
    workflow[ReducedSampleDataBySpinChannel[Up, Up]] = intensity[0]
    workflow[ReducedSampleDataBySpinChannel[Up, Down]] = intensity[1]
    workflow[ReducedSampleDataBySpinChannel[Down, Up]] = intensity[2]
    workflow[ReducedSampleDataBySpinChannel[Down, Down]] = intensity[3]
    workflow[FlipperEfficiency[Polarizer]] = FlipperEfficiency(f)

    result = np.zeros(4)
    for pola in [Up, Down]:
        for ana in [Up, Down]:
            contrib = workflow.compute(PolarizationCorrectedData[pola, ana])
            contrib = sc.concat(
                [contrib.upup, contrib.updown, contrib.downup, contrib.downdown],
                'dummy',
            )
            result += contrib.values
    np.testing.assert_allclose(result, ground_truth)


@pytest.mark.parametrize('f1', [0.1, 0.5, 0.9, 0.99, 1.0])
@pytest.mark.parametrize('f2', [0.1, 0.5, 0.9, 0.99, 1.0])
def test_workflow_with_two_flipper_computes_and_applies_matrix_inverse(
    f1: float, f2: float
) -> None:
    ground_truth = np.array([7.0, 11.0, 13.0, 17.0])
    analyzer = np.array([[1.3, 0.9], [0.9, 1.3]])
    polarizer = np.array([[1.1, 0.7], [0.7, 1.1]])
    identity = np.array([[1.0, 0.0], [0.0, 1.0]])

    flipper1 = np.array([[1.0, 0.0], [1 - f1, f1]])
    flipper2 = np.array([[1.0, 0.0], [1 - f2, f2]])
    intensity = (
        np.kron(identity, analyzer @ flipper2)
        @ np.kron(flipper1 @ polarizer, identity)
        @ ground_truth
    )

    workflow = CorrectionWorkflow()
    workflow[TransmissionFunction[Analyzer]] = FakeTransmissionFunction(analyzer)
    workflow[TransmissionFunction[Polarizer]] = FakeTransmissionFunction(polarizer)
    workflow[ReducedSampleDataBySpinChannel[Up, Up]] = intensity[0]
    workflow[ReducedSampleDataBySpinChannel[Up, Down]] = intensity[1]
    workflow[ReducedSampleDataBySpinChannel[Down, Up]] = intensity[2]
    workflow[ReducedSampleDataBySpinChannel[Down, Down]] = intensity[3]
    workflow[FlipperEfficiency[Polarizer]] = FlipperEfficiency(f1)
    workflow[FlipperEfficiency[Analyzer]] = FlipperEfficiency(f2)

    result = np.zeros(4)
    for pola in [Up, Down]:
        for ana in [Up, Down]:
            contrib = workflow.compute(PolarizationCorrectedData[pola, ana])
            contrib = sc.concat(
                [contrib.upup, contrib.updown, contrib.downup, contrib.downdown],
                'dummy',
            )
            result += contrib.values
    np.testing.assert_allclose(result, ground_truth)


@pytest.mark.parametrize('f1', [0.1, 0.99])
@pytest.mark.parametrize('f2', [0.1, 0.99])
def test_sum_contributions_provider(f1: float, f2: float) -> None:
    ground_truth = np.array([7.0, 11.0, 13.0, 17.0])
    analyzer = np.array([[1.3, 0.9], [0.9, 1.3]])
    polarizer = np.array([[1.1, 0.7], [0.7, 1.1]])
    identity = np.array([[1.0, 0.0], [0.0, 1.0]])

    flipper1 = np.array([[1.0, 0.0], [1 - f1, f1]])
    flipper2 = np.array([[1.0, 0.0], [1 - f2, f2]])
    intensity = (
        np.kron(identity, analyzer @ flipper2)
        @ np.kron(flipper1 @ polarizer, identity)
        @ ground_truth
    )

    workflow = CorrectionWorkflow()
    workflow[TransmissionFunction[Analyzer]] = FakeTransmissionFunction(analyzer)
    workflow[TransmissionFunction[Polarizer]] = FakeTransmissionFunction(polarizer)
    workflow[ReducedSampleDataBySpinChannel[Up, Up]] = intensity[0]
    workflow[ReducedSampleDataBySpinChannel[Up, Down]] = intensity[1]
    workflow[ReducedSampleDataBySpinChannel[Down, Up]] = intensity[2]
    workflow[ReducedSampleDataBySpinChannel[Down, Down]] = intensity[3]
    workflow[FlipperEfficiency[Polarizer]] = FlipperEfficiency(f1)
    workflow[FlipperEfficiency[Analyzer]] = FlipperEfficiency(f2)
    d = workflow.compute(TotalPolarizationCorrectedData)
    np.testing.assert_allclose(
        np.array([d.upup.value, d.updown.value, d.downup.value, d.downdown.value]),
        ground_truth,
    )


_he3_workflow = He3CellWorkflow()
_supermirror_workflow = SupermirrorWorkflow()
_polarizing_element_workflows = [_he3_workflow, _supermirror_workflow]


@pytest.mark.parametrize("polarizer_workflow", _polarizing_element_workflows)
@pytest.mark.parametrize("analyzer_workflow", _polarizing_element_workflows)
def test_polarization_analysis_workflow_creation(
    polarizer_workflow: sciline.Pipeline, analyzer_workflow: sciline.Pipeline
) -> None:
    workflow = PolarizationAnalysisWorkflow(
        polarizer_workflow=polarizer_workflow, analyzer_workflow=analyzer_workflow
    )
    handler = sciline.HandleAsComputeTimeException()
    for elem, wf in zip(
        (Polarizer, Analyzer), (polarizer_workflow, analyzer_workflow), strict=True
    ):
        graph = workflow.get(TransmissionFunction[elem], handler=handler)
        assert (
            pol.supermirror.SupermirrorEfficiencyFunction[elem] in graph.keys()
        ) == (wf is _supermirror_workflow)
        assert (pol.he3.He3CellPressure[elem] in graph.keys()) == (wf is _he3_workflow)


@pytest.mark.parametrize('f', [0.1, 0.5, 0.9, 0.99, 1.0])
def test_half_polarized_with_flipper_computes_and_applies_matrix_inverse(
    f: float,
) -> None:
    ground_truth = np.array([7.0, 11.0])
    polarizer = np.array([[0.96, 0.04], [0.04, 0.96]])
    flipper = np.array([[1.0, 0.0], [1 - f, f]])
    intensity = flipper @ polarizer @ ground_truth

    workflow = CorrectionWorkflow(half_polarized=True)
    workflow[TransmissionFunction[Polarizer]] = FakeTransmissionFunction(polarizer)
    workflow[ReducedSampleDataBySpinChannel[Up, pol.NoAnalyzer]] = intensity[0]
    workflow[ReducedSampleDataBySpinChannel[Down, pol.NoAnalyzer]] = intensity[1]
    workflow[FlipperEfficiency[Polarizer]] = FlipperEfficiency(f)

    result = np.zeros(2)
    for pola in [Up, Down]:
        contrib = workflow.compute(HalfPolarizedCorrectedData[pola])
        contrib = sc.concat([contrib.up, contrib.down], 'dummy')
        result += contrib.values
    np.testing.assert_allclose(result, ground_truth)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import pytest
import scipp as sc
from scipp.testing import assert_identical

from ess.polarization import he3


def test_opacity_from_cell_params() -> None:
    # In practice, the cell pressure, length, and temperature are constant, this is
    # just for testing.
    pressure = sc.array(dims=['pressure'], values=[1.0, 2.0], unit='bar')
    length = sc.array(dims=['cell_length'], values=[1.0, 2.0], unit='m')
    temperature = sc.array(dims=['temperature'], values=[200.0, 400.0], unit='K')
    wavelength = sc.array(dims=['wavelength'], values=[1.0, 2.0], unit='nm')
    opacity0 = he3.he3_opacity_from_cell_params(
        pressure=pressure, length=length, temperature=temperature
    )
    opacity_function = he3.he3_opacity_function_from_cell_opacity(opacity0)
    opacity = opacity_function(wavelength).data
    assert_identical(2 * opacity['pressure', 0], opacity['pressure', 1])
    assert_identical(2 * opacity['cell_length', 0], opacity['cell_length', 1])
    assert_identical(2 * opacity['wavelength', 0], opacity['wavelength', 1])
    assert_identical(opacity['temperature', 0], 2 * opacity['temperature', 1])
    assert opacity.unit == ''


def test_opacity_from_cell_params_reproduces_literature_value() -> None:
    # From Lee, Wai et al. (2023). Polarisation Development at the European Spallation
    # Source. EPJ Web of Conferences. 286. 10.1051/epjconf/202328603004.
    # At T = 20 deg C, p = 1 bar, lambda = 1 Angstrom, l = 1 cm we should get 0.0733.
    pressure = sc.scalar(1.0, unit='bar')
    length = sc.scalar(0.01, unit='m')
    temperature = sc.scalar(293.15, unit='K')
    wavelength = sc.scalar(1.0, unit='angstrom')
    opacity0 = he3.he3_opacity_from_cell_params(
        pressure=pressure, length=length, temperature=temperature
    )
    opacity_function = he3.he3_opacity_function_from_cell_opacity(opacity0)
    opacity = opacity_function(wavelength).data
    assert sc.isclose(opacity, sc.scalar(0.0733, unit=''), rtol=sc.scalar(1e-3))


def test_opacity_from_cell_params_raises_with_temperature_in_degree_celsius() -> None:
    pressure = sc.scalar(1.0, unit='bar')
    length = sc.scalar(1.0, unit='m')
    temperature = sc.scalar(200.0, unit='degC')
    with pytest.raises(sc.UnitError):
        he3.he3_opacity_from_cell_params(
            pressure=pressure, length=length, temperature=temperature
        )


def test_opacity_from_beam_data() -> None:
    wavelength = sc.array(dims=['wavelength'], values=[1.0, 2.0], unit='nm')
    transmission_empty_glass = sc.scalar(0.5)
    # Pretend known opacity0 for testing
    opacity0 = sc.scalar(0.3, unit='1/nm')
    ratio = transmission_empty_glass * sc.exp(-opacity0 * wavelength)
    transmission = sc.DataArray(ratio, coords={'wavelength': wavelength})
    opacity_function = he3.he3_opacity_function_from_beam_data(
        transmission_empty_glass=transmission_empty_glass,
        transmission_fraction=transmission,
        opacity0_initial_guess=opacity0 * 1.23,  # starting guess imperfect
    )
    opacity = opacity_function(wavelength).data
    assert sc.isclose(
        opacity_function.opacity0, opacity0.to(unit=opacity_function.opacity0.unit)
    )
    assert_identical(2 * opacity['wavelength', 0], opacity['wavelength', 1])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from collections.abc import Mapping
from typing import NewType

import numpy as np
import sciline as sl
import scipp as sc

from .he3 import Polarized, ReducedDirectBeamData, ReducedDirectBeamDataNoCell
from .types import (
    Analyzer,
    Down,
    Polarizer,
    PolarizingElement,
    ReducedSampleDataBySpinChannel,
    Up,
)

spin_up = sc.scalar(1, dtype='int64', unit=None)
spin_down = sc.scalar(-1, dtype='int64', unit=None)

WavelengthBins = NewType('WavelengthBins', sc.Variable)


SampleInBeamLog = NewType('SampleInBeamLog', sc.DataArray)
"""Whether the sample is in the beam as a time series."""


class CellInBeamLog(sl.Scope[PolarizingElement, sc.DataArray], sc.DataArray):
    """Whether a given cell is in the beam as a time series."""


class CellSpinLog(sl.Scope[PolarizingElement, sc.DataArray], sc.DataArray):
    """Spin state of a given cell, as a time series."""


RunSectionLog = NewType('RunSectionLog', sc.Dataset)
"""
Run section as a time series.

Derived from several time-series logs in the NeXus file, e.g.,
whether the sample and cells are in the beam or not.
"""


def determine_run_section(
    sample_in_beam: SampleInBeamLog,
    polarizer_in_beam: CellInBeamLog[Polarizer],
    analyzer_in_beam: CellInBeamLog[Analyzer],
    polarizer_spin: CellSpinLog[Polarizer],
    analyzer_spin: CellSpinLog[Analyzer],
) -> RunSectionLog:
    from scipp.scipy.interpolate import interp1d

    logs = {
        'sample_in_beam': sample_in_beam,
        'polarizer_in_beam': polarizer_in_beam,
        'analyzer_in_beam': analyzer_in_beam,
        'polarizer_spin': polarizer_spin,
        'analyzer_spin': analyzer_spin,
    }
    # TODO Change this to datetime64
    times = [
        log.coords['time'].to(unit='s', dtype='float64', copy=False)
        for log in logs.values()
    ]
    times = sc.concat(times, 'time')
    times = sc.array(dims=times.dims, unit=times.unit, values=np.unique(times.values))
    logs = {
        name: interp1d(log, 'time', kind='previous', fill_value='extrapolate')(times)
        for name, log in logs.items()
    }
    return RunSectionLog(sc.Dataset(logs))


ReducedDataByRunSectionAndWavelength = NewType(
    'ReducedDataByRunSectionAndWavelength', sc.DataArray
)


def dummy_reduction(
    time_bands: sc.Variable,
    wavelength_bands: sc.Variable,
) -> sc.DataArray:
    """This is a placeholder returning meaningless data with correct shape."""
    data = time_bands[:-1] * wavelength_bands[:-1]
    data = data / data.sum()
    return sc.DataArray(
        data, coords={'time': time_bands, 'wavelength': wavelength_bands}
    )


def run_reduction_workflow(
    run_section: RunSectionLog,
    wavelength_bands: WavelengthBins,
) -> ReducedDataByRunSectionAndWavelength:
    """
    Run the reduction workflow.

    Note that is it currently not clear if we will wrap the workflow in a function,
    or assemble a common workflow. The structural details may thus be subject to
    change.

    The reduction workflow must return normalized event data, binned into time and
    wavelength bins. The time bands define intervals of different meaning, such as
    sample runs, direct beam runs, and spin states.
    """
    # TODO
    # Subdivide sample section into smaller intervals, or return numerator/denominator
    # separately? The latter would complicate things when supporting different
    # kinds of workflows, performing different kinds of normalizations.
    # We need to be careful when subdividing and (1) exactly preserve existing bounds
    # and (2) introduce new bounds using some heuristics that yield approximately
    # equal time intervals (for the sample runs).
    data = dummy_reduction(
        time_bands=run_section.coords['time'],
        wavelength_bands=wavelength_bands,
    )
    for name, log in run_section.items():
        data.coords[name] = log.data
    return ReducedDataByRunSectionAndWavelength(data)


def extract_direct_beam(
    data: ReducedDataByRunSectionAndWavelength,
) -> ReducedDirectBeamDataNoCell:
    """Extract direct beam without any cells from direct beam data."""
    is_direct_beam = ~(
        data.coords['sample_in_beam']
        | data.coords['polarizer_in_beam']
        | data.coords['analyzer_in_beam']
    )
    # We select all bins that correspond to direct-beam run sections. This preserves
    # the separation into distinct direct beam runs, which is required later for
    # fitting a time-decay function.
    return ReducedDirectBeamDataNoCell(data[is_direct_beam])


def extract_polarizer_direct_beam_polarized(
    data: ReducedDataByRunSectionAndWavelength,
) -> ReducedDirectBeamData[Polarizer, Polarized]:
    """Extract run sections with polarized polarizer from direct beam data."""
    # TODO We need all "polarized" runs, can we assume that
    # ReducedDataByRunSectionAndWavelength does not contain any depolarized data?
    select = (
        data.coords['polarizer_in_beam']
        & ~data.coords['sample_in_beam']
        & ~data.coords['analyzer_in_beam']
    )
    return ReducedDirectBeamData[Polarizer, Polarized](data[select])


def extract_analyzer_direct_beam_polarized(
    data: ReducedDataByRunSectionAndWavelength,
) -> ReducedDirectBeamData[Analyzer, Polarized]:
    """Extract run sections with polarized analyzer from direct beam data."""
    # TODO We need all "polarized" runs, can we assume that
    # ReducedDataByRunSectionAndWavelength does not contain any depolarized data?
    select = (
        data.coords['analyzer_in_beam']
        & ~data.coords['sample_in_beam']
        & ~data.coords['polarizer_in_beam']
    )
    return ReducedDirectBeamData[Analyzer, Polarized](data[select])


def is_sample_channel(
    coords: Mapping[str, sc.Variable],
    polarizer_spin: sc.Variable,
    analyzer_spin: sc.Variable,
) -> sc.Variable:
    return (
        coords['sample_in_beam']
        & coords['polarizer_in_beam']
        & coords['analyzer_in_beam']
        & (coords['polarizer_spin'] == polarizer_spin)
        & (coords['analyzer_spin'] == analyzer_spin)
    )


def extract_sample_data_up_up(
    data: ReducedDataByRunSectionAndWavelength,
) -> ReducedSampleDataBySpinChannel[Up, Up]:
    """Extract sample data for spin channel up-up."""
    return ReducedSampleDataBySpinChannel[Up, Up](
        is_sample_channel(data, spin_up, spin_up)
    )


def extract_sample_data_up_down(
    data: ReducedDataByRunSectionAndWavelength,
) -> ReducedSampleDataBySpinChannel[Up, Down]:
    """Extract sample data for spin channel up-down."""
    return ReducedSampleDataBySpinChannel[Up, Down](
        is_sample_channel(data, spin_up, spin_down)
    )


def extract_sample_data_down_up(
    data: ReducedDataByRunSectionAndWavelength,
) -> ReducedSampleDataBySpinChannel[Down, Up]:
    """Extract sample data for spin channel down-up."""
    return ReducedSampleDataBySpinChannel[Down, Up](
        is_sample_channel(data, spin_down, spin_up)
    )


def extract_sample_data_down_down(
    data: ReducedDataByRunSectionAndWavelength,
) -> ReducedSampleDataBySpinChannel[Down, Down]:
    """Extract sample data for spin channel down-down."""
    return ReducedSampleDataBySpinChannel[Down, Down](
        is_sample_channel(data, spin_down, spin_down)
    )


providers = (
    determine_run_section,
    run_reduction_workflow,
    extract_direct_beam,
    extract_polarizer_direct_beam_polarized,
    extract_analyzer_direct_beam_polarized,
    extract_sample_data_down_down,
    extract_sample_data_down_up,
    extract_sample_data_up_down,
    extract_sample_data_up_up,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import threading
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Generic

import mantid.api as _mantid_api
import sciline as sl
import scipp as sc
from mantid import simpleapi as _mantid_simpleapi
from scippnexus import NXsource

import ess.isissans as isis
from ess.isissans.io import LoadedFileContents
from ess.isissans.mantidio import DataWorkspace, Period
from ess.reduce.nexus.types import Position
from ess.sans.types import (
    EmptyBeamRun,
    Filename,
    Incident,
    MonitorType,
    NeXusComponent,
    NeXusMonitorName,
    RunType,
    SampleRun,
    Transmission,
    TransmissionRun,
    UncertaintyBroadcastMode,
)

# In this case the "sample" is the analyzer cell, of which we want to measure
# the transmission fraction.
sample_run_type = RunType


def load_histogrammed_run(
    filename: Filename[sample_run_type], period: Period
) -> DataWorkspace[sample_run_type]:
    """Load a non-event-data ISIS file"""
    # Loading many small files with Mantid is, for some reason, very slow when using
    # the default number of threads in the Dask threaded scheduler (1 thread worked
    # best, 2 is a bit slower but still fast). We can either limit that thread count,
    # or add a lock here, which is more specific.
    with load_histogrammed_run.lock:
        loaded = _mantid_simpleapi.Load(Filename=str(filename), StoreInADS=False)
    if isinstance(loaded, _mantid_api.Workspace):
        # A single workspace
        data_ws = loaded
        if isinstance(data_ws, _mantid_api.WorkspaceGroup):
            if period is None:
                raise ValueError(
                    f'Needs {Period} to be set to know what '
                    'section of the event data to load'
                )
            data_ws = data_ws.getItem(period)
    else:
        # Separate data and monitor workspaces
        data_ws = loaded.OutputWorkspace
        if isinstance(data_ws, _mantid_api.WorkspaceGroup):
            if period is None:
                raise ValueError(
                    f'Needs {Period} to be set to know what '
                    'section of the event data to load'
                )
            data_ws = data_ws.getItem(period)
            data_ws.setMonitorWorkspace(loaded.MonitorWorkspace.getItem(period))
        else:
            data_ws.setMonitorWorkspace(loaded.MonitorWorkspace)
    return DataWorkspace[sample_run_type](data_ws)


load_histogrammed_run.lock = threading.Lock()


def _get_time(dg: sc.DataGroup) -> sc.Variable:
    start = sc.datetime(dg['run_start'].value)
    end = sc.datetime(dg['run_end'].value)
    delta = end - start
    return start + delta // 2


def _get_time_dependent_monitor(*monitor_groups: sc.DataGroup) -> sc.DataGroup:
    monitors = [grp['data'] for grp in monitor_groups]
    monitor = sc.concat(monitors, 'time')
    positions = [grp['position'] for grp in monitor_groups]
    position = _get_unique_position(*positions)
    datetime = monitor.coords['datetime']
    monitor.coords['time'] = datetime - datetime.min()
    del monitor.coords['spectrum']
    del monitor.coords['detector_id']
    return sc.DataGroup(data=monitor, position=position)


def _get_unique_position(*positions: sc.DataArray) -> sc.DataArray:
    unique = positions[0]
    for position in positions[1:]:
        if not sc.identical(position, unique):
            raise ValueError("Monitors have different source positions")
    return unique


@dataclass
class MonitorSpectrumNumber(Generic[MonitorType]):
    value: int


def get_monitor_data(
    dg: LoadedFileContents[RunType], nexus_name: NeXusMonitorName[MonitorType]
) -> NeXusComponent[MonitorType, RunType]:
    """
    Same as :py:func:`ess.isissans.get_monitor_data` but dropping variances.

    Dropping variances is a workaround required since ESSsans does not handle
    variance broadcasting when combining monitors. In our case some of the monitors
    are time-dependent, so this is required for now.
    """
    # See https://github.com/scipp/sciline/issues/52 why copy needed
    mon = dg['monitors'][nexus_name]['data'].copy()
    return NeXusComponent[MonitorType, RunType](
        sc.DataGroup(data=sc.values(mon), position=mon.coords['position'])
    )


def get_monitor_data_from_empty_beam_run(
    dg: LoadedFileContents[EmptyBeamRun],
    spectrum_number: MonitorSpectrumNumber[MonitorType],
) -> NeXusComponent[MonitorType, EmptyBeamRun]:
    """
    Extract incident or transmission monitor from ZOOM empty beam run

    The files in this case do not contain detector data, only monitor data. Mantid
    stores this as a Workspace2D, where each spectrum corresponds to a monitor.
    """
    # Note we index with a scipp.Variable, i.e., by the spectrum number used at ISIS
    monitor = sc.values(dg["data"]["spectrum", sc.index(spectrum_number.value)]).copy()
    return sc.DataGroup(data=monitor, position=monitor.coords['position'])


def get_monitor_data_from_transmission_run(
    dg: LoadedFileContents[TransmissionRun[RunType]],
    spectrum_number: MonitorSpectrumNumber[MonitorType],
) -> NeXusComponent[MonitorType, TransmissionRun[RunType]]:
    """
    Extract incident or transmission monitor from ZOOM direct-beam run

    The files in this case do not contain detector data, only monitor data. Mantid
    stores this as a Workspace2D, where each spectrum corresponds to a monitor.
    """
    # Note we index with a scipp.Variable, i.e., by the spectrum number used at ISIS
    monitor = dg['data']['spectrum', sc.index(spectrum_number.value)].copy()
    monitor.coords['datetime'] = _get_time(dg)
    return sc.DataGroup(data=monitor, position=monitor.coords['position'])


def ZoomTransmissionFractionWorkflow(runs: Sequence[str]) -> sl.Pipeline:
    """
    Workflow computing time-dependent SANS transmission fraction from ZOOM data.

    The time-dependence is obtained by using a sequence of runs.

    .. code-block:: python

        workflow = ZoomTransmissionFractionWorkflow(cell_runs)

    Note that in this case the "sample" (of which the transmission is to be computed)
    is the He3 analyzer cell.

    Parameters
    ----------
    runs:
        List of filenames of the runs to use for the transmission fraction.
    """
    workflow = isis.zoom.ZoomWorkflow()
    workflow.insert(get_monitor_data)
    workflow.insert(get_monitor_data_from_empty_beam_run)
    workflow.insert(get_monitor_data_from_transmission_run)
    workflow.insert(load_histogrammed_run)

    mapped = workflow.map({Filename[TransmissionRun[SampleRun]]: runs})
    for mon_type in (Incident, Transmission):
        workflow[NeXusComponent[mon_type, TransmissionRun[SampleRun]]] = mapped[
            NeXusComponent[mon_type, TransmissionRun[SampleRun]]
        ].reduce(func=_get_time_dependent_monitor)
        workflow[Position[NXsource, TransmissionRun[SampleRun]]] = mapped[
            Position[NXsource, TransmissionRun[SampleRun]]
        ].reduce(func=_get_unique_position)

    # We are dealing with two different types of files, and monitors are identified
    # differently in each case, so there is some duplication here.
    workflow[MonitorSpectrumNumber[Incident]] = MonitorSpectrumNumber[Incident](3)
    workflow[MonitorSpectrumNumber[Transmission]] = MonitorSpectrumNumber[Transmission](
        4
    )
    workflow[NeXusMonitorName[Incident]] = NeXusMonitorName[Incident]("monitor3")
    workflow[NeXusMonitorName[Transmission]] = NeXusMonitorName[Transmission](
        "monitor4"
    )
    workflow[UncertaintyBroadcastMode] = UncertaintyBroadcastMode.upper_bound

    return workflow



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
import numpy as np
import pytest
import scipp as sc
from scipp.testing import assert_allclose, assert_identical

from ess import polarization as pol
from ess.polarization import base

# Setup logs for four sections of length 250:
# - 10 s direct beam no cell (only in beginning)
# - For each section:
#   - 25 s direct beam with polarizer
#   - 25 s direct beam with analyzer
#   - 200 s sample run (4 states)
no_cell_start = sc.scalar(-10.0, unit='s')
section_length = sc.scalar(250.0, unit='s')


def _offset_time(da: sc.DataArray, offset: sc.Variable) -> sc.DataArray:
    da = da.copy()
    da.coords['time'] += offset
    return da


def dummy_analyzer_spin() -> base.CellSpinLog[pol.Analyzer]:
    time = sc.array(dims=['time'], values=[0.0, 150], unit='s')
    spin = sc.array(dims=['time'], values=[1, -1], unit=None)
    da = sc.DataArray(spin, coords={'time': time})
    da = sc.concat(
        [
            _offset_time(da, 0 * section_length),
            _offset_time(da, 1 * section_length),
            _offset_time(da, 2 * section_length),
            _offset_time(da, 3 * section_length),
        ],
        'time',
    )
    da.coords['time'][0] = no_cell_start
    return base.CellSpinLog[pol.Analyzer](da)


def dummy_polarizer_spin() -> base.CellSpinLog[pol.Polarizer]:
    time = sc.array(dims=['time'], values=[100.0, 200], unit='s')
    spin = sc.array(dims=['time'], values=[-1, 1], unit=None)
    da = sc.DataArray(spin, coords={'time': time})
    start = da[1].copy()
    start.coords['time'] = no_cell_start
    da = sc.concat(
        [
            start,
            _offset_time(da, 0 * section_length),
            _offset_time(da, 1 * section_length),
            _offset_time(da, 2 * section_length),
            _offset_time(da, 3 * section_length),
        ],
        'time',
    )
    return base.CellSpinLog[pol.Polarizer](da)


def dummy_sample_in_beam() -> base.SampleInBeamLog:
    time = sc.array(
        dims=['time'],
        values=[no_cell_start.value, 50, 250, 300, 500, 550, 750, 800],
        unit='s',
    )
    in_beam = sc.array(
        dims=['time'], values=[False, True, False, True, False, True, False, True]
    )
    return base.SampleInBeamLog(sc.DataArray(in_beam, coords={'time': time}))


def dummy_polarizer_in_beam() -> base.CellInBeamLog[pol.Polarizer]:
    time = sc.array(dims=['time'], values=[25.0, 50], unit='s')
    time = sc.concat(
        [
            no_cell_start,
            sc.scalar(0.0, unit='s'),
            time + 0 * section_length,
            time + 1 * section_length,
            time + 2 * section_length,
            time + 3 * section_length,
        ],
        'time',
    )
    in_beam = sc.array(dims=['time'], values=[False, True], unit=None)
    in_beam = sc.concat([in_beam] * 5, 'time')
    return base.CellInBeamLog[pol.Polarizer](
        sc.DataArray(in_beam, coords={'time': time})
    )


def dummy_analyzer_in_beam() -> base.CellInBeamLog[pol.Analyzer]:
    time = sc.array(dims=['time'], values=[0.0, 25], unit='s')
    time = sc.concat(
        [
            time + 0 * section_length,
            time + 1 * section_length,
            time + 2 * section_length,
            time + 3 * section_length,
        ],
        'time',
    )
    time[0] = no_cell_start
    in_beam = sc.array(dims=['time'], values=[False, True], unit=None)
    in_beam = sc.concat([in_beam] * 4, 'time')
    return base.CellInBeamLog[pol.Analyzer](
        sc.DataArray(in_beam, coords={'time': time})
    )


def make_events(size: int = 1000) -> sc.DataArray:
    rng = np.random.default_rng()
    time = sc.array(dims=['event'], values=rng.integers(0, 1000, size), unit='s')
    values = sc.array(dims=['event'], values=rng.uniform(0.0, 1.0, size))
    return sc.DataArray(values, coords={'time': time})


def test_determine_run_section() -> None:
    analyzer_spin = dummy_analyzer_spin()
    polarizer_spin = dummy_polarizer_spin()
    sample_in_beam = dummy_sample_in_beam()
    analyzer_in_beam = dummy_analyzer_in_beam()
    polarizer_in_beam = dummy_polarizer_in_beam()
    result = base.determine_run_section(
        sample_in_beam=sample_in_beam,
        analyzer_in_beam=analyzer_in_beam,
        polarizer_in_beam=polarizer_in_beam,
        analyzer_spin=analyzer_spin,
        polarizer_spin=polarizer_spin,
    )
    # 1 no-cell + 4 sections of (2 single-cell direct-beam + 4 sample runs)
    assert result.sizes == {'time': 1 + 4 * (2 + 4)}

    # Each section is: DB_pol, DB_ana, ++, -+, --, +-

    section = sc.array(
        dims=['time'], values=[0.0, 25.0, 50.0, 100.0, 150.0, 200.0], unit='s'
    )
    expected_time = sc.concat(
        [
            no_cell_start,
            section + 0 * section_length,
            section + 1 * section_length,
            section + 2 * section_length,
            section + 3 * section_length,
        ],
        'time',
    )
    for value in result.values():
        assert_identical(value.coords['time'], expected_time)

    section = [False, False, True, True, True, True]
    expected_sample_in_beam = sc.array(dims=['time'], values=[False] + 4 * section)
    assert_identical(result['sample_in_beam'].data, expected_sample_in_beam)

    section = [True, False, True, True, True, True]
    expected_polarizer_in_beam = sc.array(dims=['time'], values=[False] + 4 * section)
    assert_identical(result['polarizer_in_beam'].data, expected_polarizer_in_beam)

    section = [False, True, True, True, True, True]
    expected_analyzer_in_beam = sc.array(dims=['time'], values=[False] + 4 * section)
    assert_identical(result['analyzer_in_beam'].data, expected_analyzer_in_beam)

    section = [1, 1, 1, -1, -1, 1]
    expected_polarizer_spin = sc.array(
        dims=['time'], values=[1] + 4 * section, unit=None
    )
    assert_identical(result['polarizer_spin'].data, expected_polarizer_spin)

    section = [1, 1, 1, 1, -1, -1]
    expected_analyzer_spin = sc.array(
        dims=['time'], values=[1] + 4 * section, unit=None
    )
    assert_identical(result['analyzer_spin'].data, expected_analyzer_spin)


def make_IofQ(size: int = 1000) -> sc.DataArray:
    rng = np.random.default_rng()
    wavelength = sc.array(
        dims=['event'], values=rng.uniform(0.5, 5.0, size), unit='angstrom'
    )
    qx = sc.array(
        dims=['event'], values=rng.uniform(-3.1, 3.0, size), unit='1/angstrom'
    )
    qy = sc.array(
        dims=['event'], values=rng.uniform(-2.6, 2.5, size), unit='1/angstrom'
    )
    weights = sc.array(dims=['event'], values=rng.uniform(0.0, 1.0, size))
    # There are different DB runs at different times, we assume in `direct_beam` this
    # has been grouped by time already.
    time = sc.array(dims=['event'], values=rng.integers(0, 10, size))
    events = sc.DataArray(
        weights,
        coords={
            'wavelength': wavelength,
            'Qx': qx,
            'Qy': qy,
            'time': time,
        },
    )
    return events.group('time')


def test_direct_beam_returns_expected_dims() -> None:
    data = make_IofQ()
    wavelength = sc.linspace(
        dim='wavelength', start=0.5, stop=5.0, num=100, unit='angstrom'
    )
    q_range = sc.array(dims=['Q'], values=[0.0, 1.0], unit='1/angstrom')
    background_q_range = sc.array(dims=['Q'], values=[1.0, 2.0], unit='1/angstrom')

    db = pol.he3.direct_beam(
        data=data.bin(wavelength=wavelength),
        q_range=q_range,
        background_q_range=background_q_range,
    )
    assert db.bins is None
    assert db.dims == ('time', 'wavelength')


def test_direct_beam_raises_if_q_ranges_overlap() -> None:
    data = make_IofQ()
    wavelength = sc.linspace(
        dim='wavelength', start=0.5, stop=5.0, num=100, unit='angstrom'
    )
    q_range = sc.array(dims=['Q'], values=[0.0, 1.0], unit='1/angstrom')
    background_q_range = sc.array(dims=['Q'], values=[0.5, 2.0], unit='1/angstrom')

    with pytest.raises(
        ValueError, match='Background range must be after direct beam range'
    ):
        pol.he3.direct_beam(
            data=data.bin(wavelength=wavelength),
            q_range=q_range,
            background_q_range=background_q_range,
        )


def test_direct_beam_raises_if_q_range_not_positive() -> None:
    data = make_IofQ()
    wavelength = sc.linspace(
        dim='wavelength', start=0.5, stop=5.0, num=100, unit='angstrom'
    )
    q_range = sc.array(dims=['Q'], values=[-2.0, 2.0], unit='1/angstrom')
    background_q_range = sc.array(dims=['Q'], values=[3.0, 4.0], unit='1/angstrom')

    with pytest.raises(ValueError, match='Q-range must be positive'):
        pol.he3.direct_beam(
            data=data.bin(wavelength=wavelength),
            q_range=q_range,
            background_q_range=background_q_range,
        )


def test_direct_beam_operates_on_normalized_data() -> None:
    data = make_IofQ(size=int(1e6))
    data2 = data.bins.concatenate(data)
    wavelength = sc.linspace(
        dim='wavelength', start=0.5, stop=5.0, num=100, unit='angstrom'
    )
    q_range = sc.array(dims=['Q'], values=[0.0, 1.0], unit='1/angstrom')
    background_q_range = sc.array(dims=['Q'], values=[1.0, 2.0], unit='1/angstrom')

    db = pol.he3.direct_beam(
        data=data.bin(wavelength=wavelength),
        q_range=q_range,
        background_q_range=background_q_range,
    )
    db2 = pol.he3.direct_beam(
        data=data2.bin(wavelength=wavelength),
        q_range=q_range,
        background_q_range=background_q_range,
    )
    assert_allclose(db, db2)


def test_direct_beam_raises_if_input_is_counts() -> None:
    data = make_IofQ()
    wavelength = sc.linspace(
        dim='wavelength', start=0.5, stop=5.0, num=100, unit='angstrom'
    )
    q_range = sc.array(dims=['Q'], values=[0.0, 1.0], unit='1/angstrom')
    background_q_range = sc.array(dims=['Q'], values=[1.0, 2.0], unit='1/angstrom')

    data.bins.unit = 'counts'
    with pytest.raises(ValueError, match='Input data must be normalized'):
        pol.he3.direct_beam(
            data=data.bin(wavelength=wavelength),
            q_range=q_range,
            background_q_range=background_q_range,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Scipp contributors (https://github.com/scipp)
_version = "1"


def _make_pooch():
    import pooch

    return pooch.create(
        path=pooch.os_cache("ess/polarization"),
        env="ESS_AMOR_DATA_DIR",
        base_url="https://public.esss.dk/groups/scipp/ess/polarization/{version}/",
        version=_version,
        registry={
            "f_drabkin_reb.csv": "md5:cf20d53ae4af7b337d06fb84ac353994",
        },
    )


_pooch = _make_pooch()


def example_polarization_efficiency_table() -> str:
    return _pooch.fetch("f_drabkin_reb.csv")


__all__ = ["example_polarization_efficiency_table"]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from abc import ABC, abstractmethod
from dataclasses import dataclass
from io import BytesIO, StringIO
from pathlib import Path
from typing import Any, Generic

import sciline
import scipp as sc
from typing_extensions import Self

from .types import PlusMinus, PolarizingElement, TransmissionFunction


class SupermirrorEfficiencyFunction(Generic[PolarizingElement], ABC):
    """Base class for supermirror efficiency functions"""

    @abstractmethod
    def __call__(self, *, wavelength: sc.Variable) -> sc.DataArray:
        """Return the efficiency of a supermirror for a given wavelength"""


@dataclass
class SecondDegreePolynomialEfficiency(
    SupermirrorEfficiencyFunction[PolarizingElement]
):
    """
    Efficiency of a supermirror as a second-degree polynomial

    The efficiency is given by a * wavelength^2 + b * wavelength + c

    Parameters
    ----------
    a:
        Coefficient of the quadratic term, with unit of 1/angstrom^2
    b:
        Coefficient of the linear term, with unit of 1/angstrom
    c:
        Constant term, dimensionless
    """

    a: sc.Variable
    b: sc.Variable
    c: sc.Variable

    def __call__(self, *, wavelength: sc.Variable) -> sc.DataArray:
        """Return the efficiency of a supermirror for a given wavelength"""
        return (
            (self.a * wavelength**2).to(unit='', copy=False)
            + (self.b * wavelength).to(unit='', copy=False)
            + self.c.to(unit='', copy=False)
        )


@dataclass
class EfficiencyLookupTable(SupermirrorEfficiencyFunction[PolarizingElement]):
    """
    Efficiency of a supermirror as a lookup table.
    The names of the columns in the table has to be "wavelength", "efficiency".

    Parameters
    ----------
    table:
        The lookup table.
    """

    table: sc.DataArray

    def __post_init__(self):
        table = self.table if self.table.variances is None else sc.values(self.table)
        self._lut = sc.lookup(table, 'wavelength')

    def __call__(self, *, wavelength: sc.Variable) -> sc.DataArray:
        """Return the efficiency of a supermirror for a given wavelength"""
        return sc.DataArray(self._lut(wavelength), coords={'wavelength': wavelength})

    @classmethod
    def from_file(
        cls,
        path: str | Path | StringIO | BytesIO,
        wavelength_colname: str,
        efficiency_colname: str,
        wavelength_unit: sc.Unit | str = 'angstrom',
        **kwargs: Any,
    ) -> Self:
        ds = sc.io.load_csv(path, **kwargs)
        wavelength = (
            ds[wavelength_colname]
            .rename_dims({ds[wavelength_colname].dim: 'wavelength'})
            .data
        )
        wavelength.unit = wavelength_unit
        efficiency = (
            ds[efficiency_colname]
            .rename_dims({ds[efficiency_colname].dim: 'wavelength'})
            .data
        )
        return cls(sc.DataArray(efficiency, coords={'wavelength': wavelength}))


@dataclass
class SupermirrorTransmissionFunction(TransmissionFunction[PolarizingElement]):
    """Wavelength-dependent transmission of a supermirror"""

    efficiency_function: SupermirrorEfficiencyFunction

    def __call__(
        self, *, wavelength: sc.Variable, plus_minus: PlusMinus
    ) -> sc.DataArray:
        """Return the transmission fraction for a given wavelength"""
        efficiency = self.efficiency_function(wavelength=wavelength)
        if plus_minus == 'plus':
            return 0.5 * (1 + efficiency)
        else:
            return 0.5 * (1 - efficiency)

    def apply(self, data: sc.DataArray, plus_minus: PlusMinus) -> sc.DataArray:
        """Apply the transmission function to a data array"""
        return self(wavelength=data.coords['wavelength'], plus_minus=plus_minus)


def get_supermirror_transmission_function(
    efficiency_function: SupermirrorEfficiencyFunction[PolarizingElement],
) -> TransmissionFunction[PolarizingElement]:
    return SupermirrorTransmissionFunction[PolarizingElement](
        efficiency_function=efficiency_function
    )


def SupermirrorWorkflow() -> sciline.Pipeline:
    """
    Workflow for computing transmission functions for supermirror polarizing elements.
    """
    return sciline.Pipeline((get_supermirror_transmission_function,))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024 Scipp contributors (https://github.com/scipp)
import io

import pytest
import scipp as sc
from scipp.testing import assert_allclose

import ess.polarization as pol
from ess.polarization.data import example_polarization_efficiency_table


def test_SecondDegreePolynomialEfficiency_raises_if_units_incompatible():
    wav = sc.scalar(1.0, unit='m')
    eff = pol.SecondDegreePolynomialEfficiency(
        a=sc.scalar(1.0, unit='1/angstrom'),
        b=sc.scalar(1.0, unit='1/angstrom'),
        c=sc.scalar(1.0),
    )
    with pytest.raises(sc.UnitError, match=" to `dimensionless` is not valid"):
        eff(wavelength=wav)

    eff = pol.SecondDegreePolynomialEfficiency(
        a=sc.scalar(1.0, unit='1/angstrom**2'),
        b=sc.scalar(1.0, unit='1/angstrom**2'),
        c=sc.scalar(1.0),
    )
    with pytest.raises(sc.UnitError, match=" to `dimensionless` is not valid"):
        eff(wavelength=wav)

    eff = pol.SecondDegreePolynomialEfficiency(
        a=sc.scalar(1.0, unit='1/angstrom**2'),
        b=sc.scalar(1.0, unit='1/angstrom'),
        c=sc.scalar(1.0, unit='1/angstrom'),
    )
    with pytest.raises(sc.UnitError, match=" to `dimensionless` is not valid"):
        eff(wavelength=wav)

    eff = pol.SecondDegreePolynomialEfficiency(
        a=sc.scalar(1.0, unit='1/angstrom**2'),
        b=sc.scalar(1.0, unit='1/angstrom'),
        c=sc.scalar(1.0),
    )
    with pytest.raises(sc.UnitError, match=" to `dimensionless` is not valid"):
        eff(wavelength=wav / sc.scalar(1.0, unit='s'))


def test_SecondDegreePolynomialEfficiency_produces_correct_values():
    a = sc.scalar(1.0, unit='1/angstrom**2')
    b = sc.scalar(2.0, unit='1/angstrom')
    c = sc.scalar(3.0)
    f = pol.SecondDegreePolynomialEfficiency(a=a, b=b, c=c)
    assert f(wavelength=sc.scalar(0.0, unit='angstrom')) == 3.0
    assert f(wavelength=sc.scalar(1.0, unit='angstrom')) == 6.0
    assert f(wavelength=sc.scalar(2.0, unit='angstrom')) == 11.0


def test_SecondDegreePolynomialEfficiency_converts_units():
    a = sc.scalar(1.0, unit='1/angstrom**2')
    b = sc.scalar(20.0, unit='1/nm')
    c = sc.scalar(3.0)
    f = pol.SecondDegreePolynomialEfficiency(a=a, b=b, c=c)
    assert f(wavelength=sc.scalar(0.0, unit='angstrom')) == 3.0
    assert f(wavelength=sc.scalar(1.0, unit='angstrom')) == 6.0
    assert f(wavelength=sc.scalar(2.0, unit='angstrom')) == 11.0
    assert f(wavelength=sc.scalar(0.0, unit='nm')) == 3.0
    assert f(wavelength=sc.scalar(0.1, unit='nm')) == 6.0
    assert f(wavelength=sc.scalar(0.2, unit='nm')) == 11.0


def test_EfficiencyLookupTable_returns_expected_result():
    data = sc.midpoints(sc.linspace('wavelength', 0, 1, 10))
    data.variances = data.values
    tab = pol.EfficiencyLookupTable(
        sc.DataArray(
            data,
            coords={'wavelength': sc.linspace('wavelength', 0, 1, 10, unit='angstrom')},
        )
    )
    x = sc.midpoints(sc.linspace('wavelength', 0, 1, 10, unit='angstrom'))
    assert_allclose(
        tab(wavelength=x),
        sc.DataArray(sc.values(tab.table.data), coords={'wavelength': x}),
    )


def test_EfficiencyLookupTable_load_from_file():
    fname = io.StringIO('a,b,c\n1,2,3\n4,5,6')
    elt = pol.EfficiencyLookupTable.from_file(
        fname, wavelength_colname='a', efficiency_colname='b'
    )
    assert_allclose(
        sc.DataArray(
            sc.array(dims=('wavelength',), values=[2, 5]),
            coords={
                'wavelength': sc.array(
                    dims=('wavelength',), values=[1, 4], unit='angstrom'
                )
            },
        ),
        elt.table,
    )


def test_EfficiencyLookupTable_load_from_example_file():
    fname = example_polarization_efficiency_table()
    elt = pol.EfficiencyLookupTable.from_file(
        fname, wavelength_colname='# X ', efficiency_colname=' Y '
    )
    assert elt.table.size == 120
    assert elt.table.coords['wavelength'].min() == sc.scalar(3.05, unit='angstrom')
    assert elt.table.coords['wavelength'].max() == sc.scalar(14.95, unit='angstrom')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Literal, NewType, TypeVar

import sciline as sl
import scipp as sc

Up = NewType('Up', int)
Down = NewType('Down', int)
PolarizerSpin = TypeVar('PolarizerSpin', Up, Down)
AnalyzerSpin = TypeVar('AnalyzerSpin', Up, Down)
NoAnalyzer = Up
"""Indicates that element is not in use, e.g., analyzer in half-polarized case."""


class ReducedSampleDataBySpinChannel(
    sl.Scope[PolarizerSpin, AnalyzerSpin, sc.DataArray], sc.DataArray
):
    """Sample data for a given spin channel."""


Analyzer = NewType('Analyzer', str)
Polarizer = NewType('Polarizer', str)
PolarizingElement = TypeVar('PolarizingElement', Analyzer, Polarizer)

PlusMinus = Literal['plus', 'minus']


class TransmissionFunction(Generic[PolarizingElement], ABC):
    """Wavelength- and time-dependent transmission for a given cell."""

    @abstractmethod
    def apply(self, data: sc.DataArray, plus_minus: PlusMinus) -> sc.DataArray: ...


@dataclass
class PolarizingElementCorrection(
    Generic[PolarizerSpin, AnalyzerSpin, PolarizingElement]
):
    """Correction factors for polarizer or analyzer."""

    diag: sc.DataArray
    off_diag: sc.DataArray

    def get(self, *, up: bool) -> tuple[sc.DataArray, sc.DataArray]:
        """Get the correction factors for up or down spin."""
        if up:
            return self.diag, self.off_diag
        return self.off_diag, self.diag


@dataclass
class PolarizationCorrection(Generic[PolarizerSpin, AnalyzerSpin]):
    """Combined correction factors for polarizer and analyzer."""

    upup: sc.DataArray
    updown: sc.DataArray
    downup: sc.DataArray
    downdown: sc.DataArray


@dataclass
class HalfPolarizedCorrection(Generic[PolarizerSpin]):
    """Combined correction factors for half-polarized case."""

    up: sc.DataArray
    down: sc.DataArray


@dataclass
class PolarizationCorrectedData(Generic[PolarizerSpin, AnalyzerSpin]):
    """
    Polarization-corrected sample data.

    The PolarizerSpin and AnalyzerSpin type parameters refer to the measurement. The
    fields in this class give the resulting data after applying the corrections. For a
    given measurement with polarizer spin `PolarizerSpin` and analyzer spin
    `AnalyzerSpin`, there will be resulting intensity in all four output fields.
    """

    upup: sc.DataArray
    updown: sc.DataArray
    downup: sc.DataArray
    downdown: sc.DataArray


"""The sum of polarization corrected data from all flipper state channels."""
TotalPolarizationCorrectedData = NewType(
    "TotalPolarizationCorrectedData", PolarizationCorrectedData
)


@dataclass
class HalfPolarizedCorrectedData(Generic[PolarizerSpin]):
    """
    Polarization-corrected sample data for half-polarized case.

    The PolarizerSpin type parameter refers to the measurement. The fields in this class
    give the resulting data after applying the corrections. For a given measurement with
    polarizer spin `PolarizerSpin`, there will be resulting intensity in both output
    fields.
    """

    up: sc.DataArray
    down: sc.DataArray


@dataclass
class FlipperEfficiency(Generic[PolarizingElement]):
    """Efficiency of a flipper"""

    value: float
